unit ghostu;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, FileCtrl,useful;

type

  TAUnit = Class(TObject)
    Name : AnsiString;
    Path : AnsiString;
    Constructor new(const aName,aPath : string);
  end;

  TForm1 = class(TForm)
    GroupBox1: TGroupBox;
    FL: TFileListBox;
    DL: TDirectoryListBox;
    DriveComboBox1: TDriveComboBox;
    M: TListBox;
    procedure FLDblClick(Sender: TObject);
    procedure FormActivate(Sender: TObject);
  private
    procedure Analyse;
    { Private declarations }
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

Const
  COMMONUNITS = ',windows,messages,sysutils,classes,graphics,controls,forms,'
               +'dialogs,stdctrls,filectrl,buttons,db,dbgrids,dbtables,grids,'
               +'shellapi,dbctrls,extctrls,printers,forms,mask,dbierrs,'
               +'dbiprocs,dbitypes,menus,comctrls,winprocs,outline,wintypes,'
               +'useful,messagedialogbx,actnlist,';
  // *** IMPORTANT ***
  // COMMONUNITS must _start_ and _end_ with a comma

implementation

{$R *.DFM}
{$H+}

Constructor TAUnit.new(const aName,aPath : string);
begin
  inherited create;
  Name := aName;
  Path := aPath;
end;


procedure TForm1.analyse;
var
  ProjFile,unitfile:TextFile;
  pl,puc,uname,pname,u,l,tu : string;
  puList : Tlist;
  i,j,ghostcount:integer;
  ghost:boolean;
begin
  // trap for no file selected
  if fl.itemindex < 0 then begin
    alertbox('No progect file selected!');
    exit;
  end;

  setCurrentdir(ExtractFileDir(FL.Filename));
  application.processmessages;

  assignfile(ProjFile,FL.filename);
  reset(Projfile);
  m.items.clear;
  m.sorted := false;
  m.items.add('Looking for USES in '+FL.filename+'...');
  application.processmessages;
  TRY  // trap some prob reading proj file
    while not eof(ProjFile) do begin
      readln(ProjFile,pl);
      pl := trim(lowercase(pl));
      if copy(pl,1,4) = 'uses' then break;
      if copy(pl,1,5) = 'begin' then begin
        alertbox('Found BEGIN but expected USES in project file');
        exit;
      end;
    end;
    if eof(ProjFile) then begin
      alertbox('No USES clause found in project file');
      exit;
    end;
    //-----at this point we have a line that starts a uses clause
    //     so read =the whole lot in one go
    puc := pl;
    while (not eof(ProjFile)) and (pos(';',puc)=0) do begin
      readln(ProjFile,pl);
      puc := puc + ','+trim(pl);
    end;
  FINALLY
    closefile(ProjFile);
  END;


  m.items.add('Found.  Interpreting...');
  application.processmessages;
  //------ Now we have captured all the uses in the project file  (string puc)
  //       report this and then interpret into a convenient list
  puc := copy(puc,1,pos(';',puc));           // up to but excluding terminating ;
  puc := trim(substr(' '+puc,' uses ',''));  // lose the key word
  puc := substr(puc,'  ','');                // lose all spaces
  puc := substr(puc,',,',',');                // lose all spaces
  puc := substr(puc,',,',',');                // lose all spaces

  pulist := TList.create;                    // list of USES in project file
  TRY
    while length(puc)> 0 do begin
      u := trim(chop(puc,','));
      if u = '' then break;  // should never happen
      // u is now either "UNIT" or "UNIT in 'FILENAME'" or "UNIT in 'FILENAME' {FORMNAME}"
      i := pos(' in ',lowercase(u));
      if i = 0 then begin
        uname := u;
        pname := '';
      end else begin
        uname := lowercase(trim(copy(u,1,i)));
        pname := trim(copy(u,i+3,999));
        i := pos('''',pname);
        if i = 0 then begin
          alertbox('Expected opening single quote not found|'+FL.filename+'|'+uname);
          exit;
        end;
        pname := copy(pname,i+1,999);
        i := pos('''',pname);
        if i = 0 then begin
          alertbox('Expected closing single quote not found|'+FL.filename+'|'+uname);
          exit;
        end;
        pname := copy(pname,1,i-1);
      end;
      puList.add(TAUnit.new(uname,pname));
      m.items.add(uname+'  '+pname);
    end;
    m.items.add('...Done');
    application.processmessages;
    //------ we have now captured the units used in the project file
    //------ Now we have this list we need to open each one with a path
    //       and look at the uses clause in each one
    m.items.clear;
    application.processmessages;
    ghostcount := 0;
    for i := 0 to pulist.count -1 do begin
      pname := TAUnit(pulist.items[i]).path;
      if pname <> '' then begin
        if not fileexists(pname) then begin
          alertbox(pname+'|This file can''t be found');
        end else begin
          //=================================================================
          //=== Should really be a sub-routine ==============================
          //=== Analyse each unit ===========================================
          //=================================================================
          //Look for ghost references in the uses clause of unit specified in PNAME
          assignfile(unitfile,pname);
          reset(unitfile);
          u := '';
          while not eof(unitfile) do begin
            readln(unitfile,l);
            l := lowercase(trim(l));
            if pos('implementation',l)=1 then break;
            if 'uses'=copy(l,1,4) then begin
              while not eof(unitfile) do begin
                u := u +','+l;
                if pos(';',l)<>0 then break;
                readln(unitfile,l);
              end;
            end;
          end;
          //--- we now have all uses clauses in this unit


          closefile(unitfile);
          // tidy up our list
          u := copy(trim(u),6,9999);     // lose opening uses
          u := substr(u,' ','');         // lose all spaces
          u := ','+substr(u,';',',');     // add preceeding comma
          puc := substr(puc,',,',',');                // lose all spaces
          puc := substr(puc,',,',',');                // lose all spaces
          u := substr(u,',uses,',',');    // lose keywords
          u := copy(u,2,length(u)-2);     // drop bracketing commas
          //--- now nice and tidy split by commas
          while length(u)>0 do begin
            tu := trim(lowercase(chop(u,',')));
            if tu <> '' then begin   // possibly a spurious blank
              // search in the list of common units
              ghost := (pos(','+tu+',',COMMONUNITS)=0);
              if ghost then begin
                for j := 0 to pulist.count -1 do begin
                  if tu = TAUnit(pulist.items[j]).name then begin
                    ghost := false;
                    break;
                  end;
                end;
              end;
              if ghost then begin
                inc(ghostcount);
                m.items.add('Ghost:'+tu+' referenced in '+pname);
              end;
            end;
          end;
        end;
      end;  
    end;
  FINALLY;
    pulist.free;
  END;
  if ghostcount = 0 then begin
    m.items.add('NO GHOST UNITS FOUND');
  end else begin
    m.sorted := true;
  end;
end;


procedure TForm1.FLDblClick(Sender: TObject);
begin
  analyse;
end;

procedure TForm1.FormActivate(Sender: TObject);
begin
  okbox('Looks for units not listed in DPR but referenced elsewhere.'
        +'|These MIGHT be obsolete but still compiled.'
        +'|Double click on file to analyse.'
        +'|PROGRAM:'+ProgramNameAndDate
        +'|AUTHOR : Peter Fox - peter@eminent.demon.co.uk'
        +'|FREE - Hack to suit yourself.'
        +'|SOURCE - www.eminent.demon.co.uk'
        +'|NO support   NO warranty   NO commercial resale.');
end;

end.
